"""
__author__ = Hagai Hargil
"""
import numpy as np
import pandas as pd
import warnings
from numba import jit
from typing import Tuple


def verify_periodicity(tag_data: pd.Series=None, tag_freq: float = 0, binwidth: float = 0,
                       tag_pulses: int = 1) -> pd.Series:
    """
    Verify the integrity of the TAG lens pulses periodicity.
    tag_freq - Operation frequency of TAG lens.
    binwidth - Of multiscaler.
    tag_pulses - How many pulses per period the drives generates.
    :return: Sorted Series with the new pulses in the right place
    """

    period = np.ceil(1 / (tag_freq * binwidth * tag_pulses)).astype(np.int32)  # Period of TAG lens in bins
    jitter = 0.05
    allowed_noise = np.ceil(jitter * period).astype(np.uint64)

    # Iteration #0 of while loop:

    # Eliminate returns on the line:
    tag_diffs_full = tag_data.diff()
    print('The mean frequency of TAG events is {:0.2f} Hz.'.format(1 / (np.mean(tag_diffs_full) * binwidth)))
    tag_diffs = tag_diffs_full > allowed_noise
    tag_diffs[0] = True
    tag_data = tag_data.loc[tag_diffs]

    # Creation of missing pulses
    tags_diff = np.abs(tag_diffs_full - period)
    tags_diff.loc[tags_diff < allowed_noise] = 0
    missing_ticks = np.where(tags_diff != 0)[0][1:]  # returns a tuple
    changed_ticks = missing_ticks.shape[0]

    if changed_ticks > 0.2 * tag_data.shape[0]:  # Corrupted data
        warnings.warn('TAG Lens data was corrupted. Stack will be created without it.')
        return tag_data.values

    iteration_idx = 1

    # Iterate until TAG pulses are completely periodic
    while changed_ticks != 0:
        new_ticks = np.take(tag_data.values, missing_ticks[missing_ticks < tag_data.shape[0]] - 1) + period
        tag_data = tag_data.append(pd.Series(new_ticks, dtype=np.uint64), ignore_index=True)
        tag_data.sort_values(inplace=True)

        tags_diff = np.abs(tag_data.diff() - period)
        tags_diff.loc[tags_diff < allowed_noise] = 0
        missing_ticks = np.where(tags_diff != 0)[0][1:] # returns a tuple
        changed_ticks_old = changed_ticks
        changed_ticks = missing_ticks.shape[0]

        if iteration_idx > 100 or changed_ticks > changed_ticks_old:
            warnings.warn('Something is wrong with the TAG interpolation, possibly an out-of-phase lens.\n\
                          Stopping interpolation process.')
            return tag_data.values
        else:
            iteration_idx += 1

    # Add pulses that didn't exist to make sure all photons indeed have a phase
    last_pulse = tag_data.iat[-1]
    for idx in range(1, 10):
        tag_data = tag_data.append(pd.Series([last_pulse + (idx * period)], dtype=np.uint64), ignore_index=True)

    return tag_data


@jit(nopython=True, cache=True)
def numba_digitize(values: np.array, bins: np.array) -> np.array:
    """
    Numba'd version of np.digitize.
    """
    bins = np.digitize(values, bins)
    relevant_bins = bins > 0
    return bins, relevant_bins


@jit(nopython=True, cache=True)
def numba_find_phase(photons: np.array, bins: np.array, raw_tag: np.array) -> np.array:
    """
    Find the phase [0, 2pi) of the photon for each event in `photons`.
    :return: Numpy array with the size of photons containing the phases.
    """

    phase_vec = np.zeros_like(photons, dtype=np.float32)
    tag_diff = np.diff(raw_tag)

    for idx, bin in enumerate(bins) :  # values of indices that changed
        phase_vec[idx] = (photons[idx] - raw_tag[bin - 1])/tag_diff[bin - 1]

    phase_vec = np.sin(phase_vec * 2 * np.pi)

    return phase_vec


def define_phase(df_photons: pd.DataFrame=None, tag_data: pd.Series=None) -> pd.DataFrame:
    """
    Assign each data point its corresponding phase between 0 and 2pi, where max value has a phase of 2pi.
    """

    bin_idx, relevant_bins = numba_digitize(df_photons['abs_time'].values, tag_data.values)
    photons = df_photons['abs_time'].values.astype(float)
    photons[np.logical_not(relevant_bins)] = np.nan
    relevant_photons = np.compress(relevant_bins, photons)

    phase_vec = numba_find_phase(photons=relevant_photons, bins=np.compress(relevant_bins, bin_idx),
                                 raw_tag=tag_data.values)
    first_relevant_photon_idx = photons.shape[0] - relevant_photons.shape[0]
    photons[first_relevant_photon_idx:] = phase_vec

    df_photons['Phase'] = photons
    df_photons.dropna(how='any', inplace=True)

    assert df_photons['Phase'].any() >= -1
    assert df_photons['Phase'].any() <= 1

    return df_photons


def interpolate_tag(df_photons: pd.DataFrame=None, tag_data: pd.Series=None, tag_freq: float=None,
                    binwidth: float=None, tag_pulses: int=None) -> Tuple[pd.DataFrame, bool]:
    """
    If a TAG channel was defined determine each photon's phase and insert it into the main DataFrame.
    :param df_photons: DataFrame to be changed - TAG phase will be added as index
    :param tag_data: Time of TAG pulses
    :return: Original dataframe with an added index corresponding to each photon's phase
    """

    # TODO: Input verification
    # TODO: Get rid of assumption that the TAG pulse comes at a phase of 0
    # TODO: Still only a single channel of data

    tag_interp_ok = False
    tag_data = verify_periodicity(tag_data=tag_data, tag_freq=tag_freq, binwidth=binwidth, tag_pulses=tag_pulses)
    if isinstance(tag_data, pd.Series):
        df_photons = define_phase(df_photons=df_photons, tag_data=tag_data)
        tag_interp_ok = True
    elif isinstance(tag_data, np.ndarray):
        tag_interp_ok = False
        try:
            padded_tag = np.pad(tag_data, (df_photons.shape[0] - len(tag_data), 0), 'constant')
            df_photons['TAG'] = padded_tag
        except ValueError:  # more TAG pulses than events
            df_photons['TAG'] = tag_data[:df_photons.shape[0]]

    return df_photons, tag_interp_ok












# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os, sys

import mock

MOCK_MODULES = ['numpy', 'scipy', 'matplotlib', 'matplotlib.pyplot', 'pandas', 'attrs', 'cython', 'tables']
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = mock.Mock()

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.ifconfig',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]
if os.getenv('SPELLCHECK'):
    extensions += 'sphinxcontrib.spelling',
    spelling_show_suggestions = True
    spelling_lang = 'en_US'

source_suffix = '.rst'
master_doc = 'index'
project = 'PySight'
year = '2017'
author = 'Hagai Hargil'
copyright = '{0}, {1}'.format(year, author)
version = release = '0.6.12'
autodoc = True

pygments_style = 'trac'
templates_path = ['.']
extlinks = {
    'issue': ('https://github.com/HagaiHargil/python-pysight/issues/%s', '#'),
    'pr': ('https://github.com/HagaiHargil/python-pysight/pull/%s', 'PR #'),
}
html_theme = "alabaster"
html_theme_options = {
    "font_family": "Palatino, Georgia, serif",
    "font_size": "18px",
}
html_use_smartypants = True
html_last_updated_fmt = '%b %d, %Y'
html_use_index = True
html_split_index = False
html_sidebars = {
   '**': ['searchbox.html', 'globaltoc.html', 'sourcelink.html'],
}
html_short_title = '%s-%s' % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



"""
__author__ = Hagai Hargil
"""
from bokeh.plotting import output_file, show, figure
import numpy as np


def f(movie):
    # Compute
    hist, _ = np.histogram(movie.data.time_rel_pulse, bins=16)
    A0 = np.max(hist)
    K0, C0 = 1/5, A0/4

    max_idx = np.argmax(hist)
    t = np.arange(max_idx, 16)
    y = hist[max_idx:]
    A, K = fit_exp_linear(t, y, C0)

    # Bokeh setups
    output_file(r"C:\Users\Hagai\to_ori_29-8-17_flim.html", title="FLIM fits")
    fig = figure(title="Exponential fit")
    fig.line(t, y, legend=f"tau={0.8/K} ns")
    fig.vbar(x=range(16), bottom=0, width=0.5, top=hist)

    show(fig)

def fit_exp_linear(t, y, C):
    y = y - C
    y = np.log(y)
    K, A_log = np.polyfit(t, y, 1)
    A = np.exp(A_log)
    return A, -K



"""
__author__ = Hagai Hargil
"""
import attr
from attr.validators import instance_of
import numpy as np
import pandas as pd


@attr.s(slots=True)
class GatedDetection(object):
    """ Gating the unneeded photons """

    raw = attr.ib(validator=instance_of(pd.DataFrame))
    reprate = attr.ib(default=80.3e6, validator=instance_of(float))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    data = attr.ib(init=False)

    @property
    def bins_bet_pulses(self) -> int:
        return int(np.ceil(1 / (self.reprate * self.binwidth)))

    @property
    def range_length(self) -> int:
        return int(self.bins_bet_pulses / 2)

    def run(self):
        """ Main pipeline of class """
        if 'time_rel_pulse' not in self.raw.columns:
            self.data = self.raw
            return
        self.__validate_time_rel_pulse()
        hist = self.__gen_hist()
        self.__discard_events(hist=hist)

    def __gen_hist(self) -> np.array:
        """ Create a histogram of the data """

        hist, _ = np.histogram(self.raw.time_rel_pulse, bins=self.bins_bet_pulses)
        return hist

    def __validate_time_rel_pulse(self):
        """ Discard events that aren't inside the right range of allowed times from a laser pulse """

        self.raw = self.raw.loc[self.raw.time_rel_pulse >= 0]
        self.raw = self.raw.loc[self.raw.time_rel_pulse < self.bins_bet_pulses]

    def __discard_events(self, hist: np.array):
        """ Throw away unneeded events """

        peak_idx = np.argmax(hist)
        lower_range = (peak_idx - 1) % self.bins_bet_pulses  # taking into consideration photons that came
                                                       # just before the peak, mainly due to resolution considerations
        upper_range = lower_range + self.range_length
        range_after_mod = np.arange(int(lower_range), int(upper_range)) % self.bins_bet_pulses
        mask = self.raw.time_rel_pulse.isin(range_after_mod)
        self.data = self.raw.loc[mask]



# encoding: utf-8
cimport cython

cimport numpy as np
import numpy as np

cdef dict hex_to_bin_dict():

    cdef dict diction = {\
            '0': '0000',
            '1': '0001',
            '2': '0010',
            '3': '0011',
            '4': '0100',
            '5': '0101',
            '6': '0110',
            '7': '0111',
            '8': '1000',
            '9': '1001',
            'a': '1010',
            'b': '1011',
            'c': '1100',
            'd': '1101',
            'e': '1110',
            'f': '1111',
        }

    return diction


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef list iter_string_hex_to_bin(str long_str):
    cdef unsigned long long idx
    cdef dict diction
    cdef list result

    result = []
    diction = hex_to_bin_dict()

    for idx in range(len(long_str)):
        result.append(diction[long_str[idx]])

    return result


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef np.ndarray convert_bin_to_int(np.ndarray arr):
    cdef np.ndarray[np.uint64_t, ndim=1] result = np.zeros((len(arr),), dtype=np.uint64)
    cdef unsigned long long idx
    cdef unsigned long long length = len(arr)

    for idx in range(length):
        result[idx] = int(arr[idx], 2)

    return result


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef np.ndarray convert_hex_to_int(np.ndarray arr):
    cdef np.ndarray[np.uint64_t, ndim=1] result = np.zeros((len(arr),), dtype=np.uint64)
    cdef unsigned long long idx
    cdef unsigned long long length = len(arr)

    for idx in range(length):
        result[idx] = int(arr[idx], 16)

    return result


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef np.ndarray convert_hex_to_bin(np.ndarray arr):
    cdef np.ndarray[object, ndim=1] result = np.empty((len(arr),), dtype=object)
    cdef unsigned long long idx
    cdef unsigned long long length = len(arr)

    for idx in range(length):
        result[idx] = bin(int(arr[idx], 16))[2:]

    return result


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef tuple get_lost_bit_np(list list_with_lost, int step_size, unsigned long long num_of_events):
    cdef list list_of_lost_bits
    cdef np.ndarray[np.uint64_t, ndim=1] timings = np.zeros((num_of_events,), dtype=np.uint64)
    cdef unsigned long long idx, lin_idx
    cdef str str_as_hex

    list_of_lost_bits = []

    for lin_idx, idx in enumerate(range(0, len(list_with_lost), step_size)):
        list_of_lost_bits.append(list_with_lost[idx][0])

        str_as_hex = "".join(list_with_lost[idx:idx+step_size])[1:]
        timings[lin_idx] = int(str_as_hex, 16)

    return list_of_lost_bits, timings


@cython.wraparound(False)
@cython.boundscheck(False)
cpdef tuple get_lost_bit_tag(list list_with_lost, int step_size, unsigned long long num_of_events):
    cdef list list_of_lost_bits
    cdef np.ndarray[object, ndim=1] timings = np.empty((num_of_events,), dtype=object)
    cdef unsigned long long idx, lin_idx
    cdef str str_as_hex

    list_of_lost_bits = []

    for lin_idx, idx in enumerate(range(0, len(list_with_lost), step_size)):
        list_of_lost_bits.append(list_with_lost[idx][0])
        timings[lin_idx] = "".join(list_with_lost[idx:idx+step_size])[1:]

    return list_of_lost_bits, timings



"""
__author__ = Hagai Hargil
"""

import unittest
import pandas as pd
import numpy as np
from pysight.gating_tools import GatedDetection


class TestGatedDetection(unittest.TestCase):
    df = pd.DataFrame([-1, 1, 5, 7, 9, 19, 25], columns=['time_rel_pulse'],
                      dtype=np.uint64)
    base_obj = GatedDetection(raw=df)

    def test_bins_bet_pulses(self):
        inst = []
        inst.append(self.base_obj)
        inst.append(GatedDetection(raw=self.df, reprate=160.3e6))
        inst.append(GatedDetection(raw=self.df, binwidth=100e-12))

        bins_bet_pulses = [16, 8, 125]
        for bins, obj in zip(bins_bet_pulses, inst):
            self.assertEqual(bins, obj.bins_bet_pulses)

    def test_validate_time(self):
        arr = np.array([[1], [5], [7], [9]])

        obj = self.base_obj
        obj._GatedDetection__validate_time_rel_pulse()
        self.assertSequenceEqual(arr.tolist(), obj.raw.values.tolist())

    def test_validate_with_boundaries(self):
        arr = np.array([[0], [5], [15]])
        df2 = pd.DataFrame([0, 5, 15, 16], columns=['time_rel_pulse'],
                           dtype=np.uint64)
        obj = GatedDetection(raw=df2)
        obj._GatedDetection__validate_time_rel_pulse()
        self.assertSequenceEqual(arr.tolist(), obj.raw.values.tolist())

    def test_discard_events(self):
        arr = np.array([1, 2, 3, 3, 3, 4, 4, 5, 6, 11, 13, 10, 15, 15], dtype=np.uint64)
        df = pd.DataFrame(arr, columns=['time_rel_pulse'], dtype=np.uint64)
        hist, _ = np.histogram(arr, bins=range(1, 17))
        obj = GatedDetection(raw=df)
        obj._GatedDetection__discard_events(hist=hist)
        self.assertSequenceEqual(obj.data.time_rel_pulse.tolist(), arr[:9].tolist())

    def test_discard_with_wrap(self):
        arr = np.array([1, 2, 3, 3, 4, 4, 11, 13, 14, 15, 15, 15, 1], dtype=np.uint64)
        df = pd.DataFrame(arr, columns=['time_rel_pulse'], dtype=np.uint64)
        hist, _ = np.histogram(arr, bins=16)
        obj = GatedDetection(raw=df)
        obj._GatedDetection__discard_events(hist=hist)
        ans_arr = [1, 2, 3, 3, 4, 4, 14, 15, 15, 15, 1]
        self.assertSequenceEqual(obj.data.time_rel_pulse.tolist(), ans_arr)



"""
__author__ = Hagai Hargil
"""
from typing import Dict, List
import pandas as pd
import numpy as np
import warnings


def validate_line_input(dict_of_data: Dict, cols_in_data: List, num_of_lines: int=-1,
                        num_of_frames: int=-1, binwidth: float=800e-12,
                        last_event_time: int=-1, line_freq: float=7930.0, bidir: bool=False,
                        delay_between_frames: float=0.0011355, use_sweeps: bool=False,
                        max_sweep: int=65535, total_sweep_time: int=1):
    """ Verify that the .lst input of lines exists and looks fine. Create one if there's no such input. """
    if num_of_lines == -1:
        raise ValueError('No number of lines input received.')

    if num_of_frames == -1:
        raise ValueError('No number of frames received.')

    if last_event_time == -1:
        raise ValueError('No last event time received.')

    if len(cols_in_data) == 0:
        raise ValueError('No columns in data.')

    # Find the suitable case for this data regarding the line signal recorded
    type_of_line_data = match_line_data_to_case(lines=dict_of_data.get('Lines'),
                                                num_of_lines=num_of_lines,
                                                keys=list(dict_of_data.keys()),
                                                use_sweeps=use_sweeps)

    if 'corrupt' == type_of_line_data:
        # Data is corrupted, and no frame channel can help us.
        line_delta = bins_bet_lines(line_freq=line_freq, binwidth=binwidth,
                                    bidir=bidir)
        dict_of_data['Lines'] = extrapolate_line_data(last_event=last_event_time,
                                                      line_point=dict_of_data['Lines'].at[0, 'abs_time'],
                                                      num_of_lines=num_of_lines,
                                                      line_delta=line_delta,
                                                      delay_between_frames=delay_between_frames,
                                                      bidir=bidir, binwidth=binwidth,
                                                      num_of_frames=num_of_frames)
        return dict_of_data, line_delta


    elif 'rebuild' == type_of_line_data:
        # Data is corrupted, but we can rebuild lines on top of the frame channel
        line_array = create_line_array(last_event_time=last_event_time, num_of_lines=num_of_lines,
                                       num_of_frames=num_of_frames)
        dict_of_data['Lines'] = pd.DataFrame(line_array, columns=['abs_time'], dtype='uint64')
        line_delta = last_event_time / (num_of_lines * int(num_of_frames))
        return dict_of_data, line_delta

    elif 'valid' == type_of_line_data:
        # Data is valid. Check whether we need a 0-time line event
        line_delta = dict_of_data['Lines'].loc[:, 'abs_time'].diff().median()
        zeroth_line_delta = np.abs(dict_of_data['Lines'].loc[0, 'abs_time'] - line_delta)/line_delta
        if zeroth_line_delta < 0.05:
            dict_of_data['Lines'] = pd.DataFrame([[0] * len(cols_in_data)],
                                                 columns=cols_in_data,
                                                 dtype='uint64')\
                .append(dict_of_data['Lines'], ignore_index=True)
        return dict_of_data, line_delta

    elif 'from_scratch' == type_of_line_data:  # create our own line array
        line_array = create_line_array(last_event_time=last_event_time, num_of_lines=num_of_lines,
                                       num_of_frames=num_of_frames)
        dict_of_data['Lines'] = pd.DataFrame(line_array, columns=['abs_time'], dtype='uint64')
        line_delta = last_event_time/(num_of_lines * int(num_of_frames))
        return dict_of_data, line_delta

    elif 'sweeps-rebuild' == type_of_line_data:  # create our own line array with the sweeps data
        line_delta = 1 / line_freq if not bidir else 1 / (2 * line_freq)
        num_of_frames = np.ceil(last_event_time / (line_delta / binwidth * num_of_lines))
        line_array = create_line_array(last_event_time=last_event_time, num_of_lines=num_of_lines,
                                       num_of_frames=num_of_frames, line_delta=line_delta/binwidth)
        dict_of_data['Lines'] = pd.DataFrame(line_array, columns=['abs_time'], dtype='uint64')
        return dict_of_data, int(line_delta / binwidth)

    elif 'sweeps-from-scratch' == type_of_line_data:  # brand new line data
        sweep_vec = np.arange(max_sweep + 1, dtype=np.uint64)
        if len(sweep_vec) < 2:
            warnings.warn("All data was registered to a single sweep. Line data will be completely simulated.")
        else:
            dict_of_data['Lines'] = pd.DataFrame(
                sweep_vec * total_sweep_time,
                columns=['abs_time'], dtype=np.uint64)
        return dict_of_data, total_sweep_time

def validate_frame_input(dict_of_data: Dict, binwidth, cols_in_data: List, num_of_lines: int=-1,
                         last_event_time: int=-1):

    if num_of_lines == -1:
        raise ValueError('No number of lines received.')

    if last_event_time == -1:
        raise ValueError('No last event time input received.')

    if len(cols_in_data) == 0:
        raise ValueError('No columns in data.')

    if 'Frames' in dict_of_data.keys():
        dict_of_data['Frames'] = pd.DataFrame([[0] * len(cols_in_data)],
                                                     columns=cols_in_data,
                                                     dtype='uint64')\
                    .append(dict_of_data['Frames'], ignore_index=True)
    else:
        frame_array = create_frame_array(lines=dict_of_data['Lines'].loc[:, 'abs_time'],
                                         last_event_time=last_event_time,
                                         pixels=num_of_lines)
        dict_of_data['Frames'] = pd.DataFrame(frame_array, columns=['abs_time'], dtype='uint64')

    return dict_of_data


def bins_bet_lines(line_freq: float=0, binwidth: float=0,
                   bidir: bool=False) -> int:
    """
    Calculate the line difference in bins, according to the frequency
    of the line signal.
    :param line_freq: Frequency of lines in Hz. Regardless of bidirectional scanning.
    :return: int - line delta
    """
    freq_in_bins = 1/(line_freq * binwidth)
    return int(freq_in_bins / 2) if bidir else freq_in_bins


def extrapolate_line_data(last_event: int, line_point: int=0,
                          line_delta: int=0, num_of_lines: int=1,
                          delay_between_frames: float=0.0011355,
                          bidir: bool=False, binwidth: float=800e-12,
                          num_of_frames: int=1) -> pd.DataFrame:
    """
    From a single line signal extrapolate the presumed line data vector. The line frequency is doubled
    the original frequency. If needed, events will be discarded later.
    :param last_event: The last moment of the experiment
    :param line_delta: Bins between subsequent lines.
    :param line_point: Start interpolation from this point.
    :param num_of_lines: Number of lines in a frame.
    :param delay_between_frames: Time (in sec) between frames.
    :param bidir: Whether the scan was bidirectional.
    :param binwidth: Binwidth of multiscaler in seconds.
    :return: pd.DataFrame of line data
    """
    # Verify input
    if line_point / line_delta > num_of_lines:  # The first recorded line came very late
        line_point = 0

    # Create the matrix containing the duplicate frame data
    delay_between_frames_in_bins = int(delay_between_frames / binwidth)
    time_of_frame = line_delta * num_of_lines \
                    + delay_between_frames_in_bins
    num_of_frames = max(int(np.floor(last_event / time_of_frame)), 1)
    time_of_frame_mat = np.tile(time_of_frame * np.arange(num_of_frames), (num_of_lines, 1))

    # Create the matrix containing the duplicate line data
    line_vec = np.arange(start=line_point, stop=line_delta*num_of_lines, step=line_delta,
                         dtype=np.uint64)
    line_vec = np.r_[np.flip(np.arange(start=line_point, stop=0, step=-line_delta,
                               dtype=np.uint64)[1:], axis=0), line_vec]
    # Check if 0 should be included
    if line_vec[0] - line_delta == 0:
        line_vec = np.r_[0, line_vec]

    line_mat = np.tile(line_vec.reshape(len(line_vec), 1), (1, num_of_frames))

    # Add them up
    assert len(line_mat) > 0
    assert len(time_of_frame_mat) > 0
    final_mat = line_mat + time_of_frame_mat
    line_vec_final = np.ravel(final_mat, order='F')

    return pd.DataFrame(line_vec_final, columns=['abs_time'], dtype=np.uint64)


def create_frame_array(lines: pd.Series=None, last_event_time: int=None,
                       pixels: int=None) -> np.ndarray:
    """Create a pandas Series of start-of-frame times"""

    if last_event_time is None or pixels is None or lines.empty:
        raise ValueError('Wrong input detected.')

    if last_event_time <= 0:
        raise ValueError('Last event time is zero or negative.')

    num_of_recorded_lines = lines.shape[0]
    actual_num_of_frames = max(num_of_recorded_lines // pixels, 1)

    if num_of_recorded_lines < pixels:
        array_of_frames = np.linspace(start=0, stop=last_event_time, num=int(actual_num_of_frames),
                                      endpoint=False, dtype=np.uint64)
    else:
        unnecess_lines = num_of_recorded_lines % pixels
        array_of_frames = lines.iloc[0 : int(num_of_recorded_lines-unnecess_lines) : pixels]

    return np.array(array_of_frames)


def create_line_array(last_event_time: int=None, num_of_lines=None, num_of_frames=None,
                      line_delta: float=0.0) -> np.ndarray:
    """Create a pandas Series of start-of-line times"""

    if (last_event_time is None) or (num_of_lines is None) or (num_of_frames is None):
        raise ValueError('Wrong input detected.')

    if (num_of_lines <= 0) or (num_of_frames <= 0):
        raise ValueError('Number of lines and frames has to be positive.')

    if last_event_time <= 0:
        raise ValueError('Last event time is zero or negative.')

    total_lines = num_of_lines * int(num_of_frames)
    if line_delta == 0.0:
        line_array = np.arange(start=0, stop=last_event_time, step=last_event_time/total_lines, dtype=np.uint64)
    else:
        line_array = np.arange(start=0, stop=last_event_time, step=line_delta, dtype=np.uint64)
    return line_array


def validate_created_data_channels(dict_of_data: Dict):
    """
    Make sure that the dictionary that contains all data channels makes sense.
    """
    assert {'PMT1', 'Lines', 'Frames'} <= set(dict_of_data.keys())  # A is subset of B

    if dict_of_data['Frames'].shape[0] > dict_of_data['Lines'].shape[0]:  # more frames than lines
        raise UserWarning('More frames than lines, consider replacing the two.')

    try:
        if dict_of_data['TAG Lens'].shape[0] < dict_of_data['Lines'].shape[0]:
            raise UserWarning('More lines than TAG pulses, consider replacing the two.')
    except KeyError:
        pass

    try:
        if dict_of_data['Laser'].shape[0] < dict_of_data['Lines'].shape[0] or \
           dict_of_data['Laser'].shape[0] < dict_of_data['Frames'].shape[0]:
            raise UserWarning('Laser pulses channel contained less ticks than the Lines or Frames channel.')
    except KeyError:
        pass

    try:
        if dict_of_data['Laser'].shape[0] < dict_of_data['TAG Lens'].shape[0]:
            raise UserWarning('Laser pulses channel contained less ticks than the TAG lens channel.')
    except KeyError:
        pass


def validate_laser_input(pulses, laser_freq: float, binwidth: float) -> pd.Series:
    """
    Create an orderly laser pulse train.
    :param pulses:
    :param laser_freq:
    :return:
    """
    import warnings

    diffs = pulses.loc[:, 'abs_time'].diff()
    rel_idx = (diffs <= 1.05*np.ceil((1 / (laser_freq * binwidth)))) & (diffs >= 0.95*np.floor((1 / (laser_freq * binwidth))))
    pulses_final = pulses[rel_idx]  # REMINDER: Laser offset wasn't added
    if len(pulses_final) < 0.9 * len(pulses):  # TODO: If there's holdafter time, there's a good chance that the pulses
                                               # will not be periodic due to this extra time.
        warnings.warn("More than 10% of pulses were filtered due to bad timings. Make sure the laser input is fine.")

    pulses_final = pd.concat([pulses.loc[:0, :], pulses_final])  # Add back the first pulse
    pulses_final.reset_index(drop=True, inplace=True)

    return pulses_final


def rectify_photons_in_uneven_lines(df: pd.DataFrame, sorted_indices: np.array, lines: pd.Series, bidir: bool=True,
                                    phase: float=0, keep_unidir: bool=False):
    """
    "Deal" with photons in uneven lines. Unidir - if keep_unidir is false, will throw them away.
    Bidir = flips them over.
    """
    uneven_lines = np.remainder(sorted_indices, 2)
    if bidir:
        time_rel_line = pd.Series(range(df.shape[0]), dtype='int64', name='time_rel_line')
        time_rel_line.loc[uneven_lines == 0] = df.loc[uneven_lines == 0, 'time_rel_line_pre_drop'].values
        # Reverse the relative time of the photons belonging to the uneven lines,
        # by subtracting their relative time from the start time of the next line
        lines_to_subtract_from = lines.loc[sorted_indices[uneven_lines.astype(bool)] + 1].values
        events_to_subtract = df.loc[np.logical_and(uneven_lines, 1), 'abs_time'].values
        time_rel_line.iloc[uneven_lines.nonzero()[0]] = lines_to_subtract_from - events_to_subtract \
            + int(np.abs((np.sin(phase) * lines[1])))  # introduce phase delay between lines
        df.insert(loc=len(df.columns), value=time_rel_line.values, column='time_rel_line')

    if not bidir and not keep_unidir:
        df = df.iloc[uneven_lines != 1, :].copy()
        df.rename(columns={'time_rel_line_pre_drop': 'time_rel_line'}, inplace=True)

    if not bidir and keep_unidir:  # Unify the excess rows and photons in them into the previous row
        sorted_indices[np.logical_and(uneven_lines, 1)] -= 1
        df.loc['Lines'] = lines.loc[sorted_indices].values

    try:
        df.drop(['time_rel_line_pre_drop'], axis=1, inplace=True)
    except ValueError:  # column label doesn't exist
        pass
    df = df.loc[df.loc[:, 'time_rel_line'] >= 0, :]

    return df


def calc_last_event_time(dict_of_data: Dict, lines_per_frame: int=-1):
    """
    Find the last event time for the experiment. Logic as follows:
    No lines \ frames data given: Last event time is the last photon time.
    Only lines data given: The last start-of-frame time is created, and the difference between subsequent frames
    in the data is added.
    Frames data exists: The last frame time plus the difference between subsequent frames is the last event time.
    :param dict_of_data: Dictionary of data.
    :param lines_per_frame: Lines per frame.
    :return: int
    """

    # Basic assertions
    if lines_per_frame < 1:
        raise ValueError('No lines per frame value received, or value was corrupt.')

    if 'PMT1' not in dict_of_data:
        raise ValueError('No PMT1 channel in dict_of_data.')

    ##
    if 'Frames' in dict_of_data:
        last_frame_time = dict_of_data['Frames'].loc[:, 'abs_time'].iloc[-1]

        if dict_of_data['Frames'].shape[0] == 1:
            return int(2 * last_frame_time)
        else:
            frame_diff = int(dict_of_data['Frames'].loc[:, 'abs_time'].diff().mean())
            return int(last_frame_time + frame_diff)

    if 'Lines' in dict_of_data:
        num_of_lines_recorded = dict_of_data['Lines'].shape[0]
        div, mod = divmod(num_of_lines_recorded, lines_per_frame)

        if num_of_lines_recorded > lines_per_frame * (div+1):  # excessive number of lines
            last_line_of_last_frame = dict_of_data['Lines'].loc[:, 'abs_time']\
                .iloc[div * lines_per_frame - 1]
            frame_diff = dict_of_data['Lines'].loc[:, 'abs_time'].iloc[div * lines_per_frame - 1] -\
                dict_of_data['Lines'].loc[:, 'abs_time'].iloc[(div - 1) * lines_per_frame]
            return int(last_line_of_last_frame + frame_diff)

        elif mod == 0:  # number of lines contained exactly in number of lines per frame
            return int(dict_of_data['Lines'].loc[:, 'abs_time'].iloc[-1] + dict_of_data['Lines']\
                .loc[:, 'abs_time'].diff().mean())

        elif num_of_lines_recorded < lines_per_frame * (div+1):
            missing_lines = lines_per_frame - mod
            line_diff = int(dict_of_data['Lines'].loc[:, 'abs_time'].diff().mean())
            return int(dict_of_data['Lines'].loc[:, 'abs_time'].iloc[-1] +\
                ((missing_lines+1) * line_diff))

    # Just PMT data
    max_pmt1 = dict_of_data['PMT1'].loc[:, 'abs_time'].max()
    try:
        max_pmt2 = dict_of_data['PMT2'].loc[:, 'abs_time'].max()
    except KeyError:
        return max_pmt1
    else:
        return max(max_pmt1, max_pmt2)


def match_line_data_to_case(lines: pd.Series, keys: list, num_of_lines: int=512,
                            use_sweeps:bool=False) -> str:
    """
    Enumerate all possibilities of line data and choose the right option
    :param lines: Line data
    :param keys: Keys of `dict_of_data` dictionary
    :param num_of_lines: Number of lines per frame
    :param use_sweeps: Whether the sweeps have a meaning for image generation.
    :return: String of the specific case. Either: 'corrupt', 'rebuild', 'valid', 'from_scratch',
    'sweeps-rebuild' or 'sweeps-from-scratch'.
    """
    if use_sweeps and 'Lines' in keys:
        return 'sweeps-rebuild'

    if use_sweeps and not 'Lines' in keys:
        return 'sweeps-from-scratch'

    if 'Lines' in keys:
        lines = lines.loc[:, 'abs_time']
        # Verify that the input is not corrupt
        if lines.shape[0] < num_of_lines // 2:
            warnings.warn("Line data was corrupt as there were too few lines.\n"
                          "Simulating line data using GUI's parameters.")
            return 'corrupt'

        max_change_pct = lines[lines.diff().pct_change(periods=1) > 0.05]
        if len(max_change_pct) / lines.shape[0] > 0.1 and 'Frames' not in keys:
            warnings.warn("Line data was corrupt - the period didn't make sense.\n"
                          f"{len(max_change_pct)} out of {lines.shape[0]} lines were mispositioned. "
                          "Simulating line data using GUI's parameters.")
            return 'corrupt'

        elif len(max_change_pct) / lines.shape[0] > 0.1 and 'Frames' in keys:
            return 'rebuild'

        elif len(max_change_pct) / lines.shape[0] < 0.1:
            return 'valid'

    else:
        return 'from_scratch'



"""
__author__ = Hagai Hargil
"""

from unittest import TestCase
from pysight.validation_tools import *


class TestFrame(TestCase):
    """
    Tests for the validation functions
    """
    import pandas as pd
    import numpy as np

    def test_last_event_empty_dict(self):
        dict_of_data = dict()
        lines_per_frame = 1
        with self.assertRaises(ValueError):
            calc_last_event_time(dict_of_data, lines_per_frame)

    def test_last_event_wrong_lines(self):
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time'])}
        lines_per_frame = 0
        with self.assertRaises(ValueError):
            calc_last_event_time(dict_of_data, lines_per_frame)

    def test_last_event_only_pmt(self):
        dict_of_data = {'PMT1': pd.DataFrame([1, 5, 3], columns=['abs_time'])}
        lines_per_frame = 1
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         5)

    def test_last_event_with_frames(self):
        frame_data = pd.DataFrame([0, 100], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Frames': frame_data,
                        'Lines': [1, 2, 3]}
        lines_per_frame = 1
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         200)

    def test_last_event_with_single_frame(self):
        frame_data = pd.DataFrame([100], columns=['abs_time'])
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Frames': frame_data,
                        'Lines': [1, 2, 3]}
        lines_per_frame = 1
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         200)

    def test_last_event_with_lines_less_than_needed_single_frame(self):
        line_data = pd.DataFrame([0, 10, 20], columns=['abs_time'])
        lines_per_frame = 5
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         50)

    def test_last_event_with_lines_less_than_needed_more_frames(self):
        line_data = pd.DataFrame([0, 10, 20, 30, 40], columns=['abs_time'])
        lines_per_frame = 3
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         60)

    def test_last_event_with_lines_just_like_needed(self):
        line_data = pd.DataFrame([0, 10, 20], columns=['abs_time'])
        lines_per_frame = 3
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         30)

    def test_last_event_with_lines_more_than_needed(self):
        line_data = pd.DataFrame(np.arange(0, 100, 10), columns=['abs_time'])
        lines_per_frame = 3
        dict_of_data = {'PMT1': pd.DataFrame([1, 2, 3], columns=['abs_time']),
                        'Lines': line_data}
        self.assertEqual(calc_last_event_time(dict_of_data, lines_per_frame),
                         120)

    def tets_bins_bet_lines(self):
        line_freq = 10
        binwidth = 0.01
        bidir = False
        self.assertEqual(10, bins_bet_lines(line_freq=line_freq,
                                            binwidth=binwidth,
                                            bidir=bidir))

    def test_bins_bet_lines_bidir(self):
        line_freq = 10
        binwidth = 0.01
        bidir = True
        self.assertEqual(5, bins_bet_lines(line_freq=line_freq,
                                            binwidth=binwidth,
                                            bidir=bidir))

    def test_extrapolate_without_zero(self):
        line_point = 9
        last_event_time = 15
        line_delta = 2
        num_of_lines = last_event_time // line_delta
        returned_lines = extrapolate_line_data(last_event=last_event_time, line_point=line_point,
                                               line_delta=line_delta, num_of_lines=num_of_lines)
        real_lines = np.arange(1, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines['abs_time'].tolist(), real_lines.tolist())

    def test_extrapolate_from_zero(self):
        line_point = 8
        last_event_time = 15
        line_delta = 1
        num_of_lines = last_event_time // line_delta
        returned_lines = extrapolate_line_data(last_event=last_event_time, line_point=line_point,
                                               line_delta=line_delta, num_of_lines=num_of_lines)
        real_lines = np.arange(0, 15, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines['abs_time'].tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint(self):
        last_event_time = 15
        line_delta = 1
        num_of_lines = last_event_time // line_delta
        returned_lines = extrapolate_line_data(last_event=last_event_time,
                                               line_delta=line_delta, num_of_lines=num_of_lines)
        real_lines = np.arange(0, 15, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines['abs_time'].tolist(), real_lines.tolist())

    def test_extrapolate_without_timepoint_and_zero(self):
        last_event_time = 15
        line_delta = 2
        num_of_lines = last_event_time // line_delta + 1
        returned_lines = extrapolate_line_data(last_event=last_event_time,
                                               line_delta=line_delta, num_of_lines=num_of_lines)
        real_lines = np.arange(0, 15, step=2, dtype=np.uint64)
        self.assertSequenceEqual(returned_lines['abs_time'].tolist(), real_lines.tolist())



"""
__author__ = Hagai Hargil
"""
import numpy as np
import pandas as pd
import warnings
from numba import jit, uint64
from typing import Tuple, List
import attr
from attr.validators import instance_of
from itertools import chain


@attr.s(slots=True)
class TagPipeline(object):
    """ Pipeline to interpolate TAG lens pulses """
    photons = attr.ib(validator=instance_of(pd.DataFrame))
    tag_pulses = attr.ib(validator=instance_of(pd.Series))
    freq = attr.ib(default=189e3, validator=instance_of(float))  # Hz
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))  # Multiscaler binwidth
    num_of_pulses = attr.ib(default=1, validator=instance_of(int))  # Number of pulses per TAG period
    finished_pipe = attr.ib(init=False)

    def run(self):
        """ Main pipeline """
        clean_tag = self.__preserve_relevant_tag_pulses().reset_index(drop=True)
        verifier = TagPeriodVerifier(tag=clean_tag, freq=self.freq/self.num_of_pulses,
                                     binwidth=self.binwidth, first_photon=np.int64(self.first_photon),
                                     last_photon=self.last_photon)
        verifier.verify()
        if verifier.success:
            phaser = TagPhaseAllocator(photons=self.photons, tag=verifier.tag,
                                       pulses_per_period=self.num_of_pulses)
            phaser.allocate_phase()
            self.photons = phaser.photons
            self.finished_pipe = True
        else:
            self.finished_pipe = False
            try:  # Add the 'TAG' column to check data after pipeline finishes
                self.photons['TAG'] = np.pad(clean_tag, (self.photons.shape[0] - len(self.tag_pulses), 0), 'constant')
            except ValueError:  # more TAG pulses than events
                self.photons['TAG'] = self.tag_pulses[:self.photons.shape[0]]

    def __preserve_relevant_tag_pulses(self):
        """ Keep only TAG pulses that are in the timeframe of the experiment """
        relevant_tag_pulses = (self.tag_pulses >= self.first_photon) & (self.tag_pulses <= self.last_photon)
        return self.tag_pulses.loc[relevant_tag_pulses]

    @property
    def first_photon(self):
        return self.photons.abs_time.min()

    @ property
    def last_photon(self):
        return self.photons.abs_time.max()


@attr.s(slots=True)
class TagPeriodVerifier(object):
    """ Verify input to the TAG pipeline, and add missing pulses accordingly """

    tag = attr.ib(validator=instance_of(pd.Series))
    last_photon = attr.ib(validator=instance_of(np.uint64))
    freq = attr.ib(default=189e3, validator=instance_of(float))
    binwidth = attr.ib(default=800e-12, validator=instance_of(float))
    jitter = attr.ib(default=0.05, validator=instance_of(float))  # Allowed jitter of signal, between 0 - 1
    first_photon = attr.ib(default=np.int64(0), validator=instance_of(np.int64))
    allowed_corruption = attr.ib(default=0.3, validator=instance_of(float))
    success = attr.ib(init=False)

    @property
    def period(self):
        return int(np.ceil(1 / (self.freq * self.binwidth)))

    @property
    def allowed_noise(self):
        return int(np.ceil(self.jitter * self.period))

    def verify(self):
        """ Main script to verify and correct the recorded TAG lens pulses """

        # Find the borders of the disordered periods
        start_idx, end_idx = self.__obtain_start_end_idx()
        # Add \ remove TAG pulses in each period
        if isinstance(start_idx, np.ndarray) and isinstance(end_idx, np.ndarray):
            self.__fix_tag_pulses(start_idx, end_idx)
            self.__add_last_event_manually()
            self.success = True
        else:
            self.success = False

    def __obtain_start_end_idx(self) -> Tuple[np.ndarray, np.ndarray]:
        """ Create two vectors corresponding to the starts and ends of the 'wrong' periods of the TAG lens """
        diffs = self.tag.diff()
        diffs[0] = self.period
        print('The mean frequency of TAG events is {:0.2f} Hz.'.format(1 / (np.mean(diffs) * self.binwidth)))
        delta = np.abs(diffs - self.period)
        diffs[delta < self.allowed_noise] = 0  # regular events
        diffs[diffs != 0] = 1
        if np.sum(diffs)/len(diffs) > self.allowed_corruption:
            warnings.warn(f"Over {self.allowed_corruption * 100}% of TAG pulses were out-of-phase."
                          " Stopping TAG interpolation.")
            return (-1, -1)

        diff_of_diffs = diffs.diff()  # a vec containing 1 at the start
        #                               of a "bad" period, and -1 at its end
        diff_of_diffs[0] = 0
        starts_and_stops = diff_of_diffs[diff_of_diffs != 0]
        start_idx = starts_and_stops[starts_and_stops == 1].index - 1
        end_idx = starts_and_stops[starts_and_stops == -1].index - 1
        return start_idx.values, end_idx.values

    def __fix_tag_pulses(self, starts: np.ndarray, ends: np.ndarray):
        """ Iterate over the disordered periods and add or remove pulses """

        if len(starts) == 0:  # Nothing fix
            return
        period = self.period
        new_data = []
        items_to_discard = []
        start_iter_at = 0
        # If start contains a 0 - manually add TAG pulses
        if starts[0] == 0:
            start_iter_at = 1
            new_ser = pd.Series(np.arange(start=self.tag[ends[0]]-period,
                                          stop=self.first_photon-1,
                                          step=-period,
                                          dtype=np.uint64), dtype=np.uint64)
            self.tag = self.tag.append(new_ser, ignore_index=True).astype(np.uint64)
            items_to_discard.append(np.arange(starts[0], ends[0]))
        jitter = self.jitter
        new_data, returned_items_to_discard = numba_iterate_over_disordered(tag=self.tag.values,
                                                                            starts=starts[start_iter_at:],
                                                                            ends=ends[start_iter_at:],
                                                                            period=period,
                                                                            jitter=jitter)
        flattened_items_to_discard = list(chain.from_iterable(items_to_discard + returned_items_to_discard))
        self.tag.drop(flattened_items_to_discard, inplace=True)
        flattened_new_data = list(chain.from_iterable(new_data))
        self.tag = self.tag.append(pd.Series(flattened_new_data, dtype=np.uint64), ignore_index=True)\
                           .sort_values().reset_index(drop=True)
        assert self.tag.dtype == np.uint64

    def __add_last_event_manually(self):
        """ Insert a 'fake' TAG event to encapsulate the last remaining photons """
        last_tag_val = self.tag.values[-1] + self.period
        self.tag = self.tag.append(pd.Series(last_tag_val, dtype=np.uint64), ignore_index=True)
        assert self.tag.dtype == np.uint64

@attr.s(slots=True)
class TagPhaseAllocator(object):
    """ Assign a phase to each photon """
    photons = attr.ib(validator=instance_of(pd.DataFrame))
    tag = attr.ib(validator=instance_of(pd.Series))
    pulses_per_period = attr.ib(default=1, validator=instance_of(int))

    def allocate_phase(self):
        """ Using Numba functions allocate the proper phase to the photons """
        bin_idx, relevant_bins = numba_digitize(self.photons.abs_time.values,
                                                self.tag.values)

        photons = self.photons.abs_time.values.astype(float)
        photons[np.logical_not(relevant_bins)] = np.nan
        relevant_photons = np.compress(relevant_bins, photons)
        phase_vec = numba_find_phase(photons=relevant_photons, bins=np.compress(relevant_bins, bin_idx),
                                     raw_tag=self.tag.values)
        first_relevant_photon_idx = photons.shape[0] - relevant_photons.shape[0]
        photons[first_relevant_photon_idx:] = phase_vec

        self.photons['Phase'] = photons
        self.photons.dropna(how='any', inplace=True)

        assert self.photons['Phase'].any() >= -1
        assert self.photons['Phase'].any() <= 1


@jit(nopython=True, cache=True)
def numba_digitize(values: np.array, bins: np.array) -> np.array:
    """
    Numba'd version of np.digitize.
    """
    bins = np.digitize(values, bins)
    relevant_bins = bins > 0
    return bins, relevant_bins


@jit(nopython=True)
def numba_find_phase(photons: np.array, bins: np.array, raw_tag: np.array) -> np.array:
    """
    Find the phase [0, 2pi) of the photon for each event in `photons`.
    :return: Numpy array with the size of photons containing the phases.
    """

    phase_vec = np.zeros_like(photons, dtype=np.float32)
    tag_diff = np.diff(raw_tag)

    for idx, cur_bin in enumerate(bins):  # values of indices that changed
        phase_vec[idx] = (photons[idx] - raw_tag[cur_bin - 1])/tag_diff[cur_bin - 1]

    phase_vec = np.sin(phase_vec * 2 * np.pi)
    return phase_vec


@jit(cache=True)
def numba_iterate_over_disordered(tag: np.ndarray, starts: np.ndarray, ends: np.ndarray,
                                  period: int, jitter: float) -> Tuple[List, List]:
    """
    Numba'd version of the main TAG iteration.
    Currently not working in nopython due to some bugs with arange
    """
    new_data = []
    items_to_discard = []
    jitter_int = period * jitter
    row_idx = 1
    for start_idx, end_idx in zip(starts, ends):
        start_val = tag[start_idx]
        end_val = tag[end_idx]
        if np.abs(end_val - start_val) - period > jitter_int:
            l = np.arange(start=end_val - period, stop=start_val,
                          step=-period, dtype=np.uint64)
            new_data.append(l)
        items_to_discard.append(np.arange(start_idx + 1, end_idx))
        row_idx += 1
    return (new_data, items_to_discard)



"""
__author__ = Hagai Hargil
"""
import pandas as pd
import numpy as np
import attr
from typing import Dict, Tuple, List
import warnings
from pysight.apply_df_funcs import get_lost_bit_np, get_lost_bit_tag, iter_string_hex_to_bin, convert_hex_to_int, convert_hex_to_bin
from pysight.validation_tools import validate_line_input, validate_frame_input, \
    validate_laser_input, validate_created_data_channels, rectify_photons_in_uneven_lines, \
    calc_last_event_time
from attr.validators import instance_of


@attr.s(slots=True)
class Tabulate(object):
    """
    Prepare lst files so that they could be imaged using later modules.
    """
    # TODO: Variable documentation
    dict_of_inputs     = attr.ib(validator=instance_of(dict))
    data               = attr.ib(validator=instance_of(np.ndarray))
    dict_of_slices_hex = attr.ib(validator=instance_of(dict))
    dict_of_slices_bin = attr.ib()
    x_pixels           = attr.ib(default=512, validator=instance_of(int))
    y_pixels           = attr.ib(default=512, validator=instance_of(int))
    timepatch          = attr.ib(default='32', validator=instance_of(str))
    data_range         = attr.ib(default=1, validator=instance_of(int))
    is_binary          = attr.ib(default=False, validator=instance_of(bool))
    num_of_frames      = attr.ib(default=1, validator=instance_of(int))
    laser_freq         = attr.ib(default=80.3e6, validator=instance_of(float))
    binwidth           = attr.ib(default=800e-12, validator=instance_of(float))
    use_tag_bits       = attr.ib(default=False, validator=instance_of(bool))
    use_sweeps         = attr.ib(default=False, validator=instance_of(bool))
    time_after_sweep   = attr.ib(default=int(96), validator=instance_of(int))
    acq_delay          = attr.ib(default=int(0), validator=instance_of(int))
    line_freq          = attr.ib(default=7900.0, validator=instance_of(float))
    line_delta         = attr.ib(init=False)
    dict_of_data       = attr.ib(init=False)
    data_to_grab       = attr.ib(init=False)


    def run(self):
        """ Pipeline of analysis """
        if self.is_binary:
            df_after_timepatch = self.tabulate_input_binary()
        else:
            df_after_timepatch = self.tabulate_input_hex()

        print('Sorted dataframe created. Starting setting the proper data channel distribution...')
        self.dict_of_data = self.determine_data_channels(df=df_after_timepatch)

    @staticmethod
    def hex_to_bin_dict():
        """
        Create a simple dictionary that maps a hex input into a 4 letter binary output.
        :return: dict
        """
        diction = \
            {
                '0': '0000',
                '1': '0001',
                '2': '0010',
                '3': '0011',
                '4': '0100',
                '5': '0101',
                '6': '0110',
                '7': '0111',
                '8': '1000',
                '9': '1001',
                'a': '1010',
                'b': '1011',
                'c': '1100',
                'd': '1101',
                'e': '1110',
                'f': '1111',
            }
        return diction

    @property
    def total_sweep_time(self):
        return self.acq_delay + self.data_range + self.time_after_sweep

    def __allocate_data_by_channel(self, df: pd.DataFrame):
        """
        Go over the channels and find the events from that specific channel, assigning
        them to a dictionary with a suitable name.
        :param df: DataFrame with data to allocate.
        :return: Dict containing the data
        """
        dict_of_data = {}
        self.data_to_grab = ['abs_time', 'edge', 'sweep', 'tag']

        for key in self.dict_of_inputs:
            relevant_values = df.loc[df['channel'] == self.dict_of_inputs[key], self.data_to_grab]
            # NUMBA SORT NOT WORKING:
            # sorted_vals = numba_sorted(relevant_values.values)
            # dict_of_data[key] = pd.DataFrame(sorted_vals, columns=['abs_time'])
            if key in ['PMT1', 'PMT2']:
                dict_of_data[key] = relevant_values.reset_index(drop=True)
                dict_of_data[key]['Channel'] = 1 if 'PMT1' == key else 2  # Channel is the spectral channel
            else:
                dict_of_data[key] = relevant_values.sort_values(by=['abs_time']).reset_index(drop=True)

        return dict_of_data

    def determine_data_channels(self, df: pd.DataFrame=None) -> Dict:
        """ Create a dictionary that contains the data in its ordered form."""

        if df.empty:
            raise ValueError('Received dataframe was empty.')

        dict_of_data = self.__allocate_data_by_channel(df=df)

        if 'Frames' in dict_of_data:
            self.num_of_frames = dict_of_data['Frames'].shape[0] + 1  # account for first frame

        # Validations
        last_event_time = calc_last_event_time(dict_of_data=dict_of_data, lines_per_frame=self.y_pixels)
        dict_of_data, self.line_delta = validate_line_input(dict_of_data=dict_of_data, num_of_lines=self.y_pixels,
                                                            num_of_frames=self.num_of_frames, line_freq=self.line_freq,
                                                            last_event_time=last_event_time,
                                                            cols_in_data=self.data_to_grab, use_sweeps=self.use_sweeps,
                                                            max_sweep=df['sweep'].max(),
                                                            total_sweep_time=self.total_sweep_time)
        dict_of_data = validate_frame_input(dict_of_data=dict_of_data,
                                            num_of_lines=self.y_pixels, binwidth=self.binwidth,
                                            last_event_time=last_event_time,
                                            cols_in_data=self.data_to_grab)
        try:
            dict_of_data['Laser'] = validate_laser_input(dict_of_data['Laser'], laser_freq=self.laser_freq,
                                                         binwidth=self.binwidth)
        except KeyError:
            pass

        validate_created_data_channels(dict_of_data)
        return dict_of_data

    def process_chan_edge(self, struct_of_data):
        """
        Simple processing scheme for the channel and edge data.
        """
        bin_array = np.array(iter_string_hex_to_bin("".join(struct_of_data.data)))
        edge      = self.slice_string_arrays(bin_array, start=0, end=1)
        channel   = self.slice_string_arrays(bin_array, start=1, end=4)

        return edge, channel

    def tabulate_input_hex(self) -> pd.DataFrame:
        """
        Reformat the read hex data into a dataframe.
        """

        if len(self.data) == 0:
            raise IOError('List file contained zero events.')

        for key in list(self.dict_of_slices_hex.keys())[1:]:
            self.dict_of_slices_hex[key].data = self.slice_string_arrays(self.data, self.dict_of_slices_hex[key].start,
                                                                         self.dict_of_slices_hex[key].end)

        if not self.use_tag_bits:
            self.dict_of_slices_hex.pop('tag', None)

        # Channel and edge information
        edge, channel = self.process_chan_edge(self.dict_of_slices_hex.pop('chan_edge'))
        # TODO: Timepatch == '3' is not supported because of this loop.

        if self.dict_of_slices_hex['lost'] is True:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if self.dict_of_slices_hex[key].needs_bits:
                    list_with_lost = iter_string_hex_to_bin("".join(self.dict_of_slices_hex[key].data))
                    step_size = self.dict_of_slices_hex[key].end - self.dict_of_slices_hex[key].start
                    if key == 'tag':
                        list_of_losts, self.dict_of_slices_hex[key].processed = get_lost_bit_tag(list_with_lost,
                                                                                                 step_size,
                                                                                                 len(self.data))
                    else:
                        list_of_losts, self.dict_of_slices_hex[key].processed = get_lost_bit_np(list_with_lost,
                                                                                                step_size,
                                                                                                len(self.data))
                else:
                    if key == 'tag':
                        self.dict_of_slices_hex[key].processed = convert_hex_to_bin(self.dict_of_slices_hex[key].data)
                    else:
                        self.dict_of_slices_hex[key].processed = convert_hex_to_int(self.dict_of_slices_hex[key].data)
        else:
            for key in list(self.dict_of_slices_hex.keys())[1:]:
                if key == 'tag':
                    self.dict_of_slices_hex[key].processed = convert_hex_to_bin(self.dict_of_slices_hex[key].data)
                else:
                    self.dict_of_slices_hex[key].processed = convert_hex_to_int(self.dict_of_slices_hex[key].data)

        # Reformat data
        df = pd.DataFrame(channel, columns=['channel'], dtype='category')
        df['edge'] = edge
        df['edge'] = df['edge'].astype('category')

        try:
            df['tag'] = self.dict_of_slices_hex['tag'].processed
        except KeyError:
            pass

        try:
            df['lost'] = list_of_losts  # TODO: Currently the LOST bit is meaningless
            df['lost'] = df['lost'].astype('category')
        except NameError:
            pass

        df['abs_time'] = np.uint64(0)
        df['sweep'] = np.uint64(0)
        df['time_rel_sweep'] = np.uint64(0)

        if 'sweep' in self.dict_of_slices_hex:
            # Each event has an acquisition delay before the start of sweep time,
            # and has to be multiplied by the sweep number and the time-after-sweep delay
            df['abs_time'] = self.dict_of_slices_hex['time_rel_sweep'].processed + \
                ((self.dict_of_slices_hex['sweep'].processed - 1) * (self.data_range + self.time_after_sweep)) + \
                self.dict_of_slices_hex['sweep'].processed * self.acq_delay
            df['sweep'] = self.dict_of_slices_hex['sweep'].processed - 1
            df['time_rel_sweep'] = self.dict_of_slices_hex['time_rel_sweep'].processed
        else:
            df['abs_time'] = self.dict_of_slices_hex['time_rel_sweep'].processed

        # Before sorting all photons make sure that no input is missing from the user. If it's missing
        # the code will ignore this channel, but not raise an exception
        actual_data_channels = set(df['channel'].cat.categories.values)
        if actual_data_channels != set(self.dict_of_inputs.values()):
            warnings.warn("Channels that were inserted in GUI don't match actual data channels recorded. \n"
                          f"The list files contains data in the following channels: {actual_data_channels}.")

        assert np.all(df['abs_time'].values >= 0)

        return df

    @staticmethod
    def slice_string_arrays(arr: np.array, start: int, end: int) -> np.array:
        """
        Slice an array of strings efficiently.
        Based on http://stackoverflow.com/questions/39042214/how-can-i-slice-each-element-of-a-numpy-array-of-strings
        with modifications for Python 3.
        """
        b = arr.view('U1').reshape(len(arr), -1)[:, start:end]
        return np.fromstring(b.tostring(), dtype='U' + str(end - start))

    def tabulate_input_binary(self) -> pd.DataFrame:
        """
        Reformat the read binary data into a dataframe.
        """
        num_of_lines = self.data.shape[0]

        for key in self.dict_of_slices_bin:
            cur_data = self.data[:, self.dict_of_slices_bin[key].start:self.dict_of_slices_bin[key].end]
            try:
                zero_arr = np.zeros(num_of_lines, self.dict_of_slices_bin[key].cols)

            except AttributeError:  # No cols field since number of bits is a multiple of 8
                # self.dict_of_slices_bin[key].data_as_
                pass
